/*
  WHAT: A simple, unoptimized implementation of fuzzysort's algorithm
*/

// UMD (Universal Module Definition) for fuzzysort
;(function(root, factory) {
  if (typeof define === 'function' && define.amd) define([], factory)
  else if (typeof module === 'object' && module.exports) module.exports = factory()
  else root.fuzzysort = factory()
}(this, function() {

  const fuzzysort = {
    single: (search, target) => {
      const searchLower = search.toLowerCase()
      const targetLower = target.toLowerCase()


      // very basic fuzzy match; to remove non-matching targets ASAP!
      // walk through target. find sequential matches.
      // if all chars aren't found then exit
      var matchesSimple = [] // target indexes
      var searchI = 0 // where we at
      var targetI = 0 // where you at
      while(true) {
        const isMatch = searchLower[searchI] === targetLower[targetI]
        if(isMatch) {
          matchesSimple.push(targetI)

          searchI += 1
          if(searchI === search.length) break // success
        }

        targetI += 1
        if(targetI === target.length) return null // failed, we're done
      }


      // Our target string successfully matched all characters in sequence!
      // Let's try a more advanced and strict test to improve the score
      // same thing but only count it as a match if it's consecutive or a beginning character!
      var wasUpper = false
      var wasAlphanum = false
      var isConsec = false
      var successStrict = false

      var matchesStrict = []
      var searchI = 0 // where we at
      var targetI = 0 // where you at
      while(true) {

        if (targetI >= target.length) {
          // We failed to find a good spot for this search char, go back to the previous search char and force it forward
          // IE: if searching for "ab" in the target "a ab" we'll match "a" to the first "a" and be unable to find a good spot for "b"
          if (searchI === 0) break
          searchI -= 1

          const lastMatch = matchesStrict.pop()
          targetI = lastMatch + 1

          isConsec = false
          // backfill history
          const targetCode = target.charCodeAt(targetI-1)
          wasUpper = targetCode>=65&&targetCode<=90
          wasAlphanum = wasUpper || targetCode>=97&&targetCode<=122 || targetCode>=48&&targetCode<=57

        } else {
          if(!isConsec) {
            const targetCode = target.charCodeAt(targetI)
            const isUpper = targetCode>=65&&targetCode<=90
            const isAlphanum = isUpper || targetCode>=97&&targetCode<=122 || targetCode>=48&&targetCode<=57
            const isBeginning = isUpper && !wasUpper || !wasAlphanum || !isAlphanum
            wasUpper = isUpper
            wasAlphanum = isAlphanum
            if (!isBeginning) { targetI += 1; continue }
          }

          const isMatch = searchLower[searchI] === targetLower[targetI]
          if(isMatch) {
            matchesStrict.push(targetI)

            searchI += 1
            if(searchI === search.length) { successStrict = true; break }

            isConsec = true
          } else {
            isConsec = false
          }

          targetI += 1
        }
      }


      // tally up the score & keep track of matches for highlighting later
      const result = {score: 0}
      const matchesBest = successStrict ? matchesStrict : matchesSimple
      var lastTargetI = -1
      for(const targetI of matchesBest) {
        // score only goes up if they're not consecutive
        if(lastTargetI !== targetI - 1) result.score += targetI
        lastTargetI = targetI
      }
      if(!successStrict) result.score *= 1000
      result.score += target.length - search.length


      // use matchesBest to highlight matches
      result.highlighted = ''
      var matchesIndex = 0
      var opened = false
      for(var i=0; i<target.length; i++) {
        if(matchesBest[matchesIndex] === i) {
          matchesIndex += 1
          if(!opened) { result.highlighted += '<b>'; opened = true }

          if(matchesIndex === matchesBest.length) {
            result.highlighted += `${target[i]}</b>${target.substr(i+1)}`
            break
          }
        } else {
          if(opened) { result.highlighted += '</b>'; opened = false }
        }
        result.highlighted += target[i]
      }

      return result
    },
  }

  return fuzzysort
})()
